import numpy as np
import pytest

from dkist_processing_math.statistics import average_numpy_arrays


@pytest.fixture()
def int_arrays_with_large_values():
    """ Some arrays that when all added together will overflow their datatype """
    arrays = [(np.zeros((10, 10)) + 10000 + (5000 * i)).astype(">i2") for i in range(4)]
    return arrays


def test_average_numpy_arrays_wrong_shape(numpy_arrays_wrong_shape):
    """
    Given: an iterable of numpy arrays that are not all the same shape
    When: averaging arrays
    Then: an error is raised as the shapes are required to be the same
    """
    with pytest.raises(ValueError):
        average_numpy_arrays(numpy_arrays_wrong_shape)


def test_average_numpy_arrays(multiple_test_arrays):
    """
    Given: an iterable of numpy arrays that are all the same shape
    When: calculating the average
    Then: a numpy array containing the average is returned
    """
    if isinstance(multiple_test_arrays, np.ndarray):
        desired_shape = multiple_test_arrays.shape
        desired_result = multiple_test_arrays
    else:
        desired_shape = multiple_test_arrays[0].shape
        desired_result = np.mean(np.array(multiple_test_arrays), axis=0)
    result = average_numpy_arrays(multiple_test_arrays)
    assert isinstance(result, np.ndarray)
    # Dividing an ndarray by an integer is a floating point division
    # and the result is always dtype=float
    assert result.dtype == float
    assert result.shape == desired_shape
    np.testing.assert_allclose(result, desired_result)


def test_average_numpy_arrays_empty_list():
    """
    Given: an empty iterable of numpy arrays
    When: calculating the average
    Then: an error is raised
    """
    with pytest.raises(ValueError):
        average_numpy_arrays([])


def test_average_with_large_values(int_arrays_with_large_values):
    """
    Given: a list of arrays with values that would overflow their datatype if they were all summed
    When: averaging the arrays
    Then: the correct values are returned
    """
    expected = np.ones((10, 10)) * 17500
    np.testing.assert_equal(expected, average_numpy_arrays(int_arrays_with_large_values))



import numpy
import numpy as np
from skimage import filters as skif


def make_binary(data: numpy.ndarray, numotsu: int = 1) -> np.ndarray:
    """Generate a binary image from an input float or int array.

    A threshold value is chosen via Otsu's method and all values below the threshold are set to 1 while everything
    else is set to 0. Thus the result is technically an _inverse_ binary image, which is useful for isolating the
    target grid as a feature of interest.

    Parameters
    ----------
    data
        Data to convert to binary

    numotsu
        The number of times to perform thresholding using Otsu's method. Numbers larger than 1 are useful if the
        pixel distribution of the data has > 2 modes.

    Returns
    -------
    An array of 1's and 0's corresponding to the thresholded input data.
    """
    threshold = np.inf
    for i in range(numotsu):
        tmp = skif.threshold_otsu(data[data < threshold])
        frac = np.sum(data < tmp) / data.size
        # TODO: make this a parameter
        if frac < 1e-3:
            break
        threshold = tmp
    binary = np.array(data < threshold, dtype=int)
    return binary



import numpy as np

from dkist_processing_math.feature import find_px_angles


def test_find_px_angles():
    """
    Given: a single numpy array
    When: finding the most significant angles in a hough transform
    Then: return the correct peak angle
    """

    theta = np.linspace(-np.pi / 4, np.pi / 4, 1500)

    def Gaussian(x, mu, sig):
        return np.exp(-np.power(x - mu, 2.0) / (2 * np.power(sig, 2.0)))

    mu = 0.1
    sigma = 0.2
    gaussian = Gaussian(theta, mu, sigma)

    H = np.zeros((20, 1500)) + gaussian[None, :]
    desired_peaktheta = np.array(mu)

    result_peaktheta = find_px_angles(H, theta)
    assert isinstance(result_peaktheta, np.ndarray)

    np.testing.assert_array_almost_equal(result_peaktheta, desired_peaktheta)



from typing import Tuple

import numpy
import numpy as np
from skimage import transform as skit


def do_hough(
    binary: numpy.ndarray, theta_min: float, theta_max: float, numtheta: int = 1500
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Perform a Hough line transform on input data.

    Resolution in the angle dimension can be increased via numtheta.

    Parameters
    ----------
    binary
        An integer array of 1's and 0's

    theta_min
        The minimum angle to use for the hough transform fitting

    theta_max
        The maximum angle to use for the hough transform fitting

    numtheta
        The number of samples in the theta dimension of the Hough transform.

    Returns
    -------
    The 2D Hough space (i.e., accumulator) of the input image

    A 1D array of the angle values corresponding to the Hough accumulator. Angles will range from -pi/4 to 3/4 pi
    and the length will be equal to numtheta.

    A 1D array of the distance values corresponding to the Hough accumulator. This array is entirely determined
    by skimage.transform.hough_line
    """
    hough = skit.hough_line(binary, theta=np.linspace(theta_min, theta_max, numtheta))
    hough_accumulator = hough[0]
    theta = hough[1]
    rho = hough[2]

    return hough_accumulator, theta, rho



from typing import Tuple

import numpy
import numpy as np
import peakutils as pku


def find_px_angles(
    hough_accumulator: numpy.ndarray,
    theta: numpy.ndarray,
) -> Tuple[np.ndarray, np.ndarray]:
    """Find the most significant angles in a Hough transform.

    Peaks in the Hough transform a first found with a simple max filter and then refined via interpolation of the
    surrounding peak.

    Note, peak angles are not allowed to be within pi/10 of each other.

    Parameters
    ----------
    hough_accumulator
        A 2D array representing a Hough accumulator

    theta
        A 1D array of the angle values corresponding to hough_accumulator

    Returns
    -------
    peak_theta
        The most significant angles found

    idx
        The index values corresponding to the location in theta of the initial (non-interpolated) guesses for most
        significant angles.

    """
    rss = np.sqrt(np.sum(hough_accumulator ** 2, axis=0))
    # This min distance limits us to only finding the single most prominent angle
    idx = pku.indexes(rss, min_dist=theta.size)
    peak_theta = pku.interpolate(theta, rss, ind=idx)

    return peak_theta



from functools import partial
from typing import Iterable
from typing import Iterator
from typing import Union

import numpy as np
from skimage import transform as skit


def bin_arrays(
    arrays: Union[Iterable[np.ndarray], np.ndarray], bin_factors: (int, int)
) -> Iterator[np.ndarray]:
    """
    Bin an iterable of arrays by summation over an integral number of pixels in each dimension.

    Parameters
    ----------
    arrays
        The arrays to be binned

    bin_factors
        The bin factors for each axis, which must be integral divisors of each array dimension

    Returns
    -------
    An Iterator yielding the binned arrays

    """
    arrays = [arrays] if isinstance(arrays, np.ndarray) else arrays
    partial_bin = partial(_bin_array, bin_factors=bin_factors)
    return map(partial_bin, arrays)


def _bin_array(array: np.ndarray, bin_factors: (int, int)) -> np.ndarray:
    """
    Bin an array by summation over an integral number of pixels in each dimension.

    Parameters
    ----------
    array
        The array to be binned

    bin_factors
        The bin factors for each axis, which must be integral divisors of each array dimension

    Returns
    -------
    The binned array

    """
    input_shape = array.shape
    if input_shape[0] % bin_factors[0] or input_shape[1] % bin_factors[1]:
        raise ValueError(
            "Array dimensions must be an integral multiple of the binning factor\n"
            f"Shape of array = {input_shape}, binning factors = {bin_factors}"
        )
    return skit.downscale_local_mean(array, bin_factors)



import numpy as np
import pytest
from scipy import ndimage as spnd
from skimage import transform as transform

from dkist_processing_math.transform import affine_transform_arrays
from dkist_processing_math.transform import bin_arrays
from dkist_processing_math.transform import do_hough
from dkist_processing_math.transform import make_binary
from dkist_processing_math.transform import rotate_arrays_about_point
from dkist_processing_math.transform import rotation_matrix
from dkist_processing_math.transform import scale_matrix
from dkist_processing_math.transform import shear_matrix
from dkist_processing_math.transform import translate_arrays
from dkist_processing_math.transform import translation_matrix


@pytest.fixture()
def binned_arrays(numpy_arrays):
    """
    Create a list of binned arrays from numpy_arrays

    Parameters
    ----------
    numpy_arrays: List[np.ndarray]
        A list of arrays to be binned

    Returns
    -------
    List[np.ndarray]
        A list of 2x2 binned versions of the input arrays

    """

    temp_arrays = np.zeros((2, 2, 2), dtype=numpy_arrays[0].dtype)
    for array, temp_array in zip(numpy_arrays, temp_arrays):
        for row in range(temp_array.shape[1]):
            for column in range(temp_array.shape[0]):
                temp_array[column, row] = array[
                    column * 5 : (column + 1) * 5, row * 5 : (row + 1) * 5
                ].mean()
    return temp_arrays


@pytest.fixture()
def binned_array(numpy_array, binned_arrays):
    """
    Create a binned array from numpy_array

    Parameters
    ----------
    numpy_array: np.ndarray
        The array to be binned

    Returns
    -------
    np.ndarray
        A 2x2 binned version of the input array

    """
    return binned_arrays[0]


def test_bin_arrays(numpy_arrays, binned_arrays):
    """
    Given: an iterable of numpy_arrays
    When: binning an iterable of numpy_arrays
    Then: an Iterator of binned versions of the input arrays is returned
    """
    results = bin_arrays(numpy_arrays, (5, 5))
    for result, binned_array in zip(results, binned_arrays):
        np.testing.assert_allclose(result, binned_array)


def test_bin_arrays_single_array(numpy_array, binned_array):
    """
    Given: an single numpy_array
    When: binning an iterable of numpy_arrays
    Then: an Iterator containing the binned version of the input arrays is returned
    """
    result = next(bin_arrays(numpy_array, (5, 5)))
    np.testing.assert_allclose(result, binned_array)


def test_bin_arrays_bad_bin_factor(numpy_array):
    """
    Given: a single numpy array
    When: binning an array using a bad binning factor
    Then: an error is raised
    """
    with pytest.raises(ValueError):
        next(bin_arrays(numpy_array, (3, 3)))


def test_affine_transform_specify_matrix(numpy_array):
    """
    Given: a numpy_array
    When: applying the affine transform and specifying the full transformation matrix
    Then: an Iterator containing the transformed array is returned
    """
    angle = np.radians(90)
    # Create the Affine transform for the rotation
    rotation = transform.AffineTransform(rotation=angle)
    # Offset to the center of the array for the rotation
    p = np.array([4.5, 4.5])
    # Create the Affine transform for the translation
    trans = transform.AffineTransform(translation=p)
    # Create the composite transform matrix for the rotation about the center
    tform = trans.params @ rotation.params @ np.linalg.inv(trans.params)
    # Perform the transformation
    result = affine_transform_arrays(numpy_array, matrix=tform)
    # Compare against the same rotation computed with a different method.
    np.testing.assert_allclose(next(result), transform.rotate(numpy_array, -90, center=(4.5, 4.5)))


def test_affine_transform_specify_rotate(numpy_array):
    """
    Given: a numpy_array
    When: applying the affine transform and specifying a rotation angle and rotation point
    Then: an Iterator containing the rotated array is returned
    """
    angle = np.radians(90)
    rotation = transform.AffineTransform(rotation=angle)
    # Offset to the center of the array for the rotation
    p = np.array([4.5, 4.5, 0])
    # Transform the offset to rotated space
    o = p - np.dot(rotation.params, p)
    result = affine_transform_arrays(numpy_array, rotation=np.radians(90), translation=o[0:2][::-1])
    np.testing.assert_allclose(next(result), spnd.rotate(numpy_array, -90))


def test_affine_transform_specify_scale_x(numpy_array):
    """
    Given: a numpy_array
    When: applying the affine transform and specifying a scale factor for the X axis only
    Then: an Iterator containing the array scaled along the X axis is returned
    """

    num_rows = numpy_array.shape[0]
    num_cols = numpy_array.shape[1]
    desired_output = np.zeros_like(numpy_array)
    desired_output[0:num_rows, 0 : num_cols // 2] = numpy_array[0:num_rows, 0:num_cols:2]
    result = affine_transform_arrays(numpy_array, scale=(0.5, 1.0))
    np.testing.assert_allclose(next(result), desired_output)


def test_affine_transform_specify_scale_y(numpy_array):
    """
    Given: a numpy_array
    When: applying the affine transform and specifying a scale factor for the Y axis only
    Then: an Iterator containing the array scaled along the Y axis is returned
    """
    num_rows = numpy_array.shape[0]
    num_cols = numpy_array.shape[1]
    desired_output = np.zeros_like(numpy_array)
    desired_output[0 : num_rows // 2, 0:num_cols] = numpy_array[0:num_rows:2, 0:num_cols]
    result = affine_transform_arrays(numpy_array, scale=(1.0, 0.5))
    np.testing.assert_allclose(next(result), desired_output)


def test_affine_transform_specify_scale_both(numpy_array):
    """
    Given: a numpy_array
    When: applying the affine transform and specifying a scale factor for each axis
    Then: an Iterator containing the array scaled separately along each axis is returned
    """
    tform = transform.SimilarityTransform(scale=0.5)
    desired_output = transform.warp(numpy_array, tform.inverse)
    result = affine_transform_arrays(numpy_array, scale=(0.5, 0.5))
    np.testing.assert_allclose(next(result), desired_output)


def test_affine_transform_specify_scale_scalar(numpy_array):
    """
    Given: a numpy_array
    When: applying the affine transform and specifying a scalar scale factor to be applied to both axes
    Then: an Iterator containing the array scaled identically on each axis is returned
    """
    tform = transform.SimilarityTransform(scale=0.5)
    desired_output = transform.warp(numpy_array, tform.inverse)
    result = affine_transform_arrays(numpy_array, scale=0.5)
    np.testing.assert_allclose(next(result), desired_output)


def test_affine_transform_specify_shear(numpy_array):
    """
    Given: a numpy_array
    When: applying the affine transform and specifying a shear angle
    Then: an Iterator containing the sheared array is returned
    """
    shear_mat = np.array([[1, 0.1, 0], [0.2, 1, 0], [0, 0, 1]])
    shear_mat_inv = np.linalg.inv(shear_mat)
    desired_output = transform.warp(numpy_array, shear_mat_inv)
    result = affine_transform_arrays(numpy_array, shear=(0.1, 0.2))
    np.testing.assert_allclose(next(result), desired_output)


def test_affine_transform_specify_translate(numpy_array):
    """
    Given: a numpy_array
    When: applying the affine transform and specifying a translation vector
    Then: an Iterator containing the translated array is returned
    """
    result = affine_transform_arrays(numpy_array, translation=(2, 5))
    desired_output = spnd.shift(numpy_array, (2, 5))
    np.testing.assert_allclose(next(result), desired_output)


def test_affine_transform_all_params_none(numpy_array):
    """
    Given: a numpy_array
    When: applying the affine transform and specifying no parameters
    Then: an error is raised
    """
    with pytest.raises(ValueError):
        affine_transform_arrays(numpy_array)


def test_affine_transform_matrix_and_one_param(numpy_array):
    """
    Given: a numpy_array
    When: applying the affine transform and specifying both a full matrix and one or more transform parameters
    Then: an error is raised
    """
    with pytest.raises(ValueError):
        affine_transform_arrays(numpy_array, matrix=np.ones((2, 2)), shear=0.1)


def test_affine_transform_matrix_wrong_shape(numpy_array):
    """
    Given: a numpy_array
    When: applying the affine transform and specifying both a full matrix and one or more transform parameters
    Then: an error is raised
    """
    with pytest.raises(ValueError):
        affine_transform_arrays(numpy_array, matrix=np.ones((2, 2)))


def test_affine_transform_arrays(numpy_arrays):
    """
    Given: an iterable of numpy_arrays
    When: applying the affine transform
    Then: a Iterator containing the transformed arrays is returned
    """
    results = affine_transform_arrays(numpy_arrays, translation=(7, 3))
    desired_outputs = [spnd.shift(array, (7, 3)) for array in numpy_arrays]
    for result, desired_output in zip(results, desired_outputs):
        np.testing.assert_allclose(result, desired_output)


def test_affine_transform_arrays_single_array(numpy_array):
    """
    Given: a single numpy array
    When: applying the affine transform to an iterable of arrays
    Then: an Iterator containing the transformed array is returned
    """
    results = affine_transform_arrays(numpy_array, translation=(2, 3))
    result = next(results)
    desired_output = spnd.shift(numpy_array, (2, 3))
    np.testing.assert_allclose(result, desired_output)


def test_rotate_arrays_about_point(numpy_arrays):
    """
    Given: an iterable of numpy arrays
    When: rotating an iterable of arrays about a point
    Then: an Iterator containing the rotated arrays is returned
    """
    angle = -45
    p = np.array([5, 2])
    results = rotate_arrays_about_point(numpy_arrays, angle=np.radians(angle), point=p)
    desired_outputs = [transform.rotate(array, angle, center=p) for array in numpy_arrays]
    for result, desired_output in zip(results, desired_outputs):
        np.testing.assert_allclose(result, desired_output)


def test_rotate_arrays_about_point_single_array(numpy_array):
    """
    Given: a single numpy array
    When: rotating an iterable of arrays about a point
    Then: an Iterator containing the rotated array is returned
    """
    angle = -45
    p = np.array([5, 2])
    results = rotate_arrays_about_point(numpy_array, angle=np.radians(angle), point=p)
    result = next(results)
    desired_output = transform.rotate(numpy_array, angle, center=p)
    np.testing.assert_allclose(result, desired_output)


def test_translate_arrays(numpy_arrays):
    """
    Given: an iterable of numpy arrays
    When: translating an iterable of arrays by a vector
    Then: an Iterator containing the translated arrays is returned
    """
    results = translate_arrays(numpy_arrays, translation=(5, 2))
    desired_outputs = [spnd.shift(array, (5, 2)) for array in numpy_arrays]
    for result, desired_output in zip(results, desired_outputs):
        np.testing.assert_allclose(result, desired_output)


def test_translate_arrays_single_array(numpy_array):
    """
    Given: a single numpy array
    When: translating an iterable of arrays by a vector
    Then: an Iterator containing the translated array is returned
    """
    results = translate_arrays(numpy_array, translation=(5, 2))
    result = next(results)
    desired_output = spnd.shift(numpy_array, (5, 2))
    np.testing.assert_allclose(result, desired_output)


def test_scale_matrix(numpy_array):
    """
    Given: a single numpy array
    When: creating a scale matrix operator
    Then: a correct scale matrix is produced
    """
    scale_mat = scale_matrix((0.2, 0.4))
    desired_output = affine_transform_arrays(numpy_array, scale=(0.2, 0.4))
    result = affine_transform_arrays(numpy_array, matrix=scale_mat)
    np.testing.assert_allclose(next(result), next(desired_output))


def test_rotation_matrix(numpy_array):
    """
    Given: a single numpy array
    When: creating a rotation matrix operator
    Then: a correct rotation matrix is produced
    """
    angle = np.radians(30)
    rotation_mat = rotation_matrix(angle)
    desired_output = affine_transform_arrays(numpy_array, rotation=angle)
    result = affine_transform_arrays(numpy_array, matrix=rotation_mat)
    np.testing.assert_allclose(next(result), next(desired_output))


def test_shear_matrix(numpy_array):
    """
    Given: a single numpy array
    When: creating a shear matrix operator
    Then: a correct shear matrix is produced
    """
    shear = (0.1, 0.2)
    shear_mat = shear_matrix(shear)
    desired_shear_mat = np.array([[1, 0.1, 0], [0.2, 1, 0], [0, 0, 1]])
    np.testing.assert_allclose(shear_mat, desired_shear_mat)
    desired_output = affine_transform_arrays(numpy_array, shear=shear)
    result = affine_transform_arrays(numpy_array, matrix=shear_mat)
    np.testing.assert_allclose(next(result), next(desired_output))


def test_shear_matrix_scalar(numpy_array):
    """
    Given: a single numpy array
    When: creating a shear matrix operator
    Then: a correct shear matrix is produced
    """
    shear = 0.1
    shear_mat = shear_matrix(shear)
    desired_shear_mat = np.array([[1, 0.1, 0], [0.1, 1, 0], [0, 0, 1]])
    np.testing.assert_allclose(shear_mat, desired_shear_mat)
    desired_output = affine_transform_arrays(numpy_array, shear=shear)
    result = affine_transform_arrays(numpy_array, matrix=shear_mat)
    np.testing.assert_allclose(next(result), next(desired_output))


def test_translation_matrix(numpy_array):
    """
    Given: a single numpy array
    When: creating a translation matrix operator
    Then: a correct translation matrix is produced
    """
    trans = (2.3, 5.7)
    trans_mat = translation_matrix(trans)
    desired_output = affine_transform_arrays(numpy_array, translation=trans)
    result = affine_transform_arrays(numpy_array, matrix=trans_mat)
    np.testing.assert_allclose(next(result), next(desired_output))


def test_translation_matrix_scalar(numpy_array):
    """
    Given: a single numpy array
    When: creating a translation matrix operator
    Then: a correct translation matrix is produced
    """
    trans = 4.2
    trans_mat = translation_matrix(trans)
    desired_output = affine_transform_arrays(numpy_array, translation=trans)
    result = affine_transform_arrays(numpy_array, matrix=trans_mat)
    np.testing.assert_allclose(next(result), next(desired_output))


def test_matrix_compositions(numpy_array):
    """
    Given: a single numpy array
    When: creating a composition of transform matrices
    Then: a correct transform results from the composition
    """
    trans = (2.3, 5.7)
    trans_mat = translation_matrix(trans)
    shear = (0.1, 0.2)
    shear_mat = shear_matrix(shear)
    angle = np.radians(30)
    rotation_mat = rotation_matrix(angle)
    scale = (0.2, 0.4)
    scale_mat = scale_matrix(scale)
    # The order in which the transforms are applied reads right to left:
    composition = trans_mat @ rotation_mat @ scale_mat @ shear_mat
    result = affine_transform_arrays(numpy_array, matrix=composition)
    desired_output = affine_transform_arrays(
        numpy_array, scale=scale, shear=shear, rotation=angle, translation=trans
    )
    np.testing.assert_allclose(next(result), next(desired_output))


def test_make_binary_image():
    """
    Given: a single numpy array
    When: generating a binary image for values in the array below a threshold
    Then: a correct binary array results from the generation
    """
    input_array = np.ones((10, 10), dtype=float)
    input_array[:5, :] += 1.0
    result = make_binary(input_array)  # Something ridiculous
    assert isinstance(result, np.ndarray)

    expected = np.ones((10, 10), dtype=int)
    expected[:5, :] = 0
    np.testing.assert_equal(result, expected)


def test_make_binary_trigger_cutoff():
    """
    Given: a single numpy array
    When: calling make_binary in such a way that the result will converge to all 0s or all 1s
    Then: make_binary returns before a single value array is created
    """
    rng = np.random.default_rng()
    input_array = rng.standard_normal((100, 100))
    result = make_binary(input_array, numotsu=100)
    frac = np.sum(result) / result.size
    assert frac >= 1e-3


def test_do_hough(numpy_array):
    """
    Given: a single numpy array
    When: performing a hough line transformation
    Then: return the correct hough space, array of angle values, and array of distance values
    """
    binary = make_binary(numpy_array)

    # Rho and Theta ranges
    thetas = np.linspace(-np.pi / 4, np.pi / 4, 1500)
    width, height = binary.shape
    diag_len = np.ceil(np.sqrt(width * width + height * height))  # max_dist
    diag_len = int(diag_len)
    num_rhos = 2 * diag_len + 1
    rhos = np.linspace(-diag_len, diag_len, num_rhos)
    # Cache some reusable values
    cos_t = np.cos(thetas)
    sin_t = np.sin(thetas)
    num_thetas = len(thetas)
    # Hough accumulator array of theta vs rho
    accumulator = np.zeros((num_rhos, num_thetas), dtype=np.uint64)
    y_idxs, x_idxs = np.nonzero(binary)  # (row, col) indexes to edges
    for i in range(len(x_idxs)):
        x = x_idxs[i]
        y = y_idxs[i]
        for t_idx in range(num_thetas):
            # Calculate rho. diag_len is added for a positive index
            rho = round(x * cos_t[t_idx] + y * sin_t[t_idx]) + diag_len
            accumulator[rho, t_idx] += 1
    desired_H = accumulator
    desired_thetas = thetas
    desired_rhos = rhos
    assert isinstance(desired_H, np.ndarray)
    assert isinstance(desired_thetas, np.ndarray)
    assert isinstance(desired_rhos, np.ndarray)

    result_H, result_theta, result_rho = do_hough(binary, theta_min=-np.pi / 4, theta_max=np.pi / 4)
    assert isinstance(result_H, np.ndarray)
    assert isinstance(result_theta, np.ndarray)
    assert isinstance(result_rho, np.ndarray)

    np.testing.assert_allclose(result_H, desired_H)
    np.testing.assert_allclose(result_theta, desired_thetas)
    np.testing.assert_allclose(result_rho, desired_rhos)



from typing import Generator

import numpy as np
import pytest
from astropy.io import fits
from dkist_processing_common.models.fits_access import FitsAccessBase

from dkist_processing_math.arithmetic import divide_arrays_by_array
from dkist_processing_math.arithmetic import divide_fits_access_by_array
from dkist_processing_math.arithmetic import subtract_array_from_arrays
from dkist_processing_math.arithmetic import subtract_array_from_fits_access


@pytest.fixture(params=["multiple", "single"])
def multiple_test_fits_access(request, numpy_arrays):

    access_list = []
    for i, array in enumerate(numpy_arrays):
        hdu = fits.PrimaryHDU(data=array)
        hdu.header["TEST"] = i
        access_list.append(FitsAccessBase(hdu=hdu))
    if request.param == "multiple":
        return access_list
    return access_list[0]


def test_subtract_array_from_arrays(multiple_test_arrays, numpy_array):
    """
    Given: an iterable of numpy arrays that are all the same shape
    When: subtracting a fixed array from each array in the iterable
    Then: an Generator of subtracted arrays is returned
    """
    if isinstance(multiple_test_arrays, np.ndarray):
        desired_shape = multiple_test_arrays.shape
        desired_result = [multiple_test_arrays - numpy_array]
    else:
        desired_shape = multiple_test_arrays[0].shape
        desired_result = multiple_test_arrays - numpy_array
    result = subtract_array_from_arrays(multiple_test_arrays, numpy_array)
    assert isinstance(result, Generator)
    for result_array, test_array in zip(result, desired_result):
        assert result_array.shape == desired_shape
        assert result_array.dtype == np.result_type(test_array)
        np.testing.assert_allclose(result_array, test_array)


def test_subtract_array_from_fits_access(multiple_test_fits_access, numpy_array):
    """
    Given: an iterable of or single FitsAccess object(s) with data
    When: subtracting an array from all input objects
    Then: a generator of FitsAccess objects is returned with correctly subtracted data
    """
    if isinstance(multiple_test_fits_access, FitsAccessBase):
        desired_result = [multiple_test_fits_access.data - numpy_array]
        desired_header_val = [multiple_test_fits_access.header["TEST"]]
    else:
        desired_result = [i.data - numpy_array for i in multiple_test_fits_access]
        desired_header_val = [i.header["TEST"] for i in multiple_test_fits_access]

    result = subtract_array_from_fits_access(multiple_test_fits_access, numpy_array)
    assert isinstance(result, Generator)

    for result_obj, test_array, test_header_val in zip(result, desired_result, desired_header_val):
        assert result_obj.data.dtype == np.result_type(test_array)
        assert result_obj.header["TEST"] == test_header_val
        np.testing.assert_allclose(result_obj.data, test_array)


def test_subtract_array_from_arrays_wrong_shape(numpy_arrays_wrong_shape, numpy_array):
    """
    Given: an iterable of numpy arrays that are not all the same shape
    When: subtracting arrays
    Then: an error is raised as the shapes are required to be the same
    """
    with pytest.raises(ValueError):
        list(subtract_array_from_arrays(numpy_arrays_wrong_shape, numpy_array))


def test_divide_arrays_by_array(multiple_test_arrays, numpy_array):
    """
    Given: an iterable of numpy arrays that are all the same shape
    When: dividing each array in the iterable by a fixed array
    Then: an Generator of divided arrays is returned
    """
    if isinstance(multiple_test_arrays, np.ndarray):
        desired_shape = multiple_test_arrays.shape
        desired_result = [multiple_test_arrays / numpy_array]
    else:
        desired_shape = multiple_test_arrays[0].shape
        desired_result = multiple_test_arrays / numpy_array
    result = divide_arrays_by_array(multiple_test_arrays, numpy_array)
    assert isinstance(result, Generator)
    for result_array, test_array in zip(result, desired_result):
        assert result_array.shape == desired_shape
        assert result_array.dtype == np.result_type(test_array)
        np.testing.assert_allclose(result_array, test_array)


def test_divide_fits_access_by_array(multiple_test_fits_access, numpy_array):
    """
    Given: an iterable of or single FitsAccess object(s) with data
    When: dividing an array from all input objects
    Then: a generator of FitsAccess objects is returned with correctly subtracted data
    """
    if isinstance(multiple_test_fits_access, FitsAccessBase):
        desired_result = [multiple_test_fits_access.data / numpy_array]
        desired_header_val = [multiple_test_fits_access.header["TEST"]]
    else:
        desired_result = [i.data / numpy_array for i in multiple_test_fits_access]
        desired_header_val = [i.header["TEST"] for i in multiple_test_fits_access]

    result = divide_fits_access_by_array(multiple_test_fits_access, numpy_array)
    assert isinstance(result, Generator)

    for result_obj, test_array, test_header_val in zip(result, desired_result, desired_header_val):
        assert result_obj.data.dtype == np.result_type(test_array)
        assert result_obj.header["TEST"] == test_header_val
        np.testing.assert_allclose(result_obj.data, test_array)


def test_divide_arrays_by_array_wrong_shape(numpy_arrays_wrong_shape, numpy_array):
    """
    Given: an iterable of numpy arrays that are not all the same shape
    When: dividing arrays
    Then: an error is raised as the shapes are required to be the same
    """
    with pytest.raises(ValueError):
        list(divide_arrays_by_array(numpy_arrays_wrong_shape, numpy_array))



import numpy as np
import pytest

rng = np.random.default_rng()


@pytest.fixture()
def numpy_arrays():
    """
    Create an iterable of random 10 x 10 arrays

    Returns
    -------
    List[np.ndarray]

    """
    arrays = [rng.standard_normal((10, 10)), rng.standard_normal((10, 10))]
    return arrays


@pytest.fixture()
def numpy_array(numpy_arrays):
    """
    Create a single random 10 x 10 array

    Returns
    -------
    np.ndarray

    """
    return numpy_arrays[0]


@pytest.fixture()
def numpy_arrays_wrong_shape():
    """
    Create an iterable of random arrays of differing sizes

    Returns
    -------
    List[np.ndarray]

    """
    arrays = [rng.standard_normal((10, 10)), rng.standard_normal((10, 15))]
    return arrays


@pytest.fixture(params=["multiple", "single"])
def multiple_test_arrays(request, numpy_arrays):
    """
    Create an input test array or list of test arrays

    Parameters
    ----------
    request: Union['multiple, 'single']
        indicates whether to return a single array or an iterable of arrays
    numpy_arrays: List[np.ndarray]
        the arrays from which the return value is selected

    Returns
    -------
    Union[List[np.ndarray], np.ndarray]

    """
    if request.param == "multiple":
        return numpy_arrays
    else:
        return numpy_arrays[0]



from typing import Iterable
from typing import Union

import numpy as np


def average_numpy_arrays(arrays: Union[Iterable[np.ndarray], np.ndarray]) -> np.ndarray:
    """
    Given an iterable of numpy arrays, calculate the pixel-wise average and return
    it in a numpy array. This will work for a single array as well, just in case...

    Parameters
    ----------
    arrays
        The arrays to be averaged

    Returns
    -------
    The average of the input arrays

    """
    arrays = [arrays] if isinstance(arrays, np.ndarray) else arrays
    count = 0  # This statement is here only to suppress an uninitialized variable warning
    output = None
    for count, array in enumerate(arrays):
        if output is None:
            output = np.array(array).astype(float)
        else:
            if array.shape != output.shape:
                raise ValueError(
                    f"All arrays must be the same shape. "
                    f"Shape of initial array = {output.shape} "
                    f"Shape of current array = {array.shape}"
                )
            output += array
    if output is not None:
        return output / (count + 1)
    raise ValueError("data_arrays is empty")



from functools import partial
from typing import Iterable
from typing import Iterator
from typing import Tuple
from typing import Union

import numpy as np
from skimage import transform as skit


def affine_transform_arrays(
    arrays: Union[Iterable[np.ndarray], np.ndarray],
    matrix: np.ndarray = None,
    scale: Union[Tuple[float, float], np.ndarray, float] = None,
    translation: Union[Tuple[float, float], np.ndarray, float] = None,
    rotation: float = None,
    shear: Union[Tuple[float, float], np.ndarray, float] = None,
    **kwargs,
) -> Iterator[np.ndarray]:
    """
    Transform an iterable of input arrays using a generalized affine transform operator.
    A transform matrix may be specified or separate parameters for scale, translation,
    rotation and shear may be specified instead. This method abstracts the use of the
    scikit-image affine transform, which may easily be replaced with one from scipy or
    another origin. If transform parameters are specified, they are applied in this order:
    shear, scale, rotation, translation

    Parameters
    ----------
    arrays
        The array(s) to be transformed

    matrix
        The transformation matrix to be used for the transform. If specified, none of
        [translation, rotation, or shear] may be used. Optional.

    Optional arguments
        If matrix is not specified, at least one of the following must be used:

            scale
                The scale factor to be applied in the transform: (Sx, Sy). If a scalar is used, the same
                value is applied to each axis.

            translation
                The translation to be applied in the transform: (Tx, Ty) in pixel units.

            rotation
                The rotation angle, in radians, to be applied to the transformation.
                A positive angle is counter-clockwise in a right handed coordinate system

            shear:
                The shear factor to be applied in the transform: (Shx, Shy). If a scalar is used, the same
                value is applied to each axis.

            **kwargs
                Optional arguments to be passed on to skimage.transform.warp() if desired. See
                https://scikit-image.org/docs/stable/api/skimage.transform.html#skimage.transform.warp
                for more details.

    Returns
    -------
    The transformed array(s)

    """
    tform = None
    any_param_set = any(param is not None for param in [scale, translation, rotation, shear])

    # We do our own pre-checks here so that we're not totally tied to use scikit-image in the future
    #
    if not any_param_set and matrix is None:
        raise ValueError(
            "You must specify matrix or at least one of scale, rotate, shear, or shift."
        )

    elif any_param_set and matrix is not None:
        raise ValueError(
            "You cannot specify both matrix and any of scale, rotate, shear, or shift."
        )

    elif matrix is not None:
        if matrix.shape != (3, 3):
            raise ValueError(f"Invalid shape of transformation matrix: {matrix.shape}")
        tform = skit.AffineTransform(matrix=matrix)

    elif any_param_set:
        # Construct the matrix from parameters
        if scale is None:
            scale = (1.0, 1.0)
        elif np.isscalar(scale):
            scale = (scale, scale)

        if translation is None:
            translation = (0.0, 0.0)
        elif np.isscalar(translation):
            translation = (translation, translation)

        if rotation is None:
            rotation = 0.0

        if shear is None:
            shear = (0.0, 0.0)

        translation = translation[::-1]  # Because skimage and numpy have different axes ordering
        shear_mat = shear_matrix(shear)
        temp_tform = skit.AffineTransform(scale=scale, translation=translation, rotation=rotation)
        tform_mat = temp_tform.params @ shear_mat
        tform = skit.AffineTransform(matrix=tform_mat)

    arrays = [arrays] if isinstance(arrays, np.ndarray) else arrays
    partial_warp = partial(skit.warp, inverse_map=tform.inverse, **kwargs)
    return map(partial_warp, arrays)


def rotate_arrays_about_point(
    arrays: Union[Iterable[np.ndarray], np.ndarray],
    angle: float = 0.0,
    point: Union[Tuple[float, float], np.ndarray] = None,
    **kwargs,
) -> Iterator[np.ndarray]:
    """
    Rotate an iterable of arrays about a specified point

    Parameters
    ----------
    arrays
        The array(s) to be rotated

    angle
        The angle, in radians, for the rotation to be applied, optional. Default is zero
        A positive angle is counter-clockwise in a right handed coordinate system and
        clockwise in a left-handed coordinate system

    point
        The point, in pixel coordinates (x, y) about which to rotate. Optional. Default is None,
        meaning the center of the array will be used. Note that the origin pixel is centered at
        (0, 0) and has extent: [-0.5: 0.5, -0.5: 0.5]

    kwargs
        Any arguments to be passed to skimage.transform.rotate. See
        https://scikit-image.org/docs/stable/api/skimage.transform.html#skimage.transform.rotate
        for more details.

    Returns
    -------
    An Iterator yielding the rotated arrays.

    """
    arrays = [arrays] if isinstance(arrays, np.ndarray) else arrays
    partial_rotate = partial(skit.rotate, angle=np.degrees(angle), center=point, **kwargs)
    return map(partial_rotate, arrays)


def translate_arrays(
    arrays: Union[Iterable[np.ndarray], np.ndarray],
    translation: Union[Tuple[float, float], np.ndarray] = (0.0, 0.0),
    **kwargs,
) -> Iterator[np.ndarray]:
    """
    Translate an iterable of arrays by a specified vector

    Parameters
    ----------
    arrays
        The array(s) to be translated

    translation
        The translation to be applied in the transform: (Tx, Ty) in pixel units. Optional.

    **kwargs
        Optional arguments to be passed on to skimage.transform.warp() if desired. See
        https://scikit-image.org/docs/stable/api/skimage.transform.html#skimage.transform.warp
        for more details.

    Returns
    -------
    np.ndarray
        The translated array

    """
    arrays = [arrays] if isinstance(arrays, np.ndarray) else arrays
    return affine_transform_arrays(arrays, translation=translation, **kwargs)


def scale_matrix(scale_factors: Union[Tuple[float, float], float] = (1.0, 1.0)) -> np.ndarray:
    """
    Create an Affine Transform matrix for a scale operation

    Parameters
    ----------
    scale_factors
        The scale factors to use in scaling the array. If a scalar is passed, it
        is applied to both axes.

    Returns
    -------
    A 3x3 homogeneous transform matrix implementing the scaling
    """
    return skit.AffineTransform(scale=scale_factors).params


def rotation_matrix(angle: float = 0.0) -> np.ndarray:
    """
    Create an Affine Transform matrix for a rotation operation
    This assumes the rotation is about (0, 0) and it is up to the user
    to calculate the appropriate translation offset to be applied in the
    skimage.warp() method to achieve rotation about an arbitrary point

    Parameters
    ----------
    angle
        The angle of rotation in radians. This represents counter-clockwise rotation
        in a right-handed coordinate system and clockwise rotation in a left-handed
        system

    Returns
    -------
    A 3x3 homogeneous transform matrix implementing the rotation
    """
    return skit.AffineTransform(rotation=angle).params


def shear_matrix(shear_factors: Union[Tuple[float, float], float] = (1.0, 1.0)) -> np.ndarray:
    """
    Create an Affine Transform matrix for a shear operation

    Parameters
    ----------
    shear_factors
        The shear scaling factors.

    Returns
    -------
    A 3x3 homogeneous transform matrix implementing the shear
    """
    if isinstance(shear_factors, float):
        shear_factors = shear_factors, shear_factors
    shear_mat = np.eye(3, 3)
    shear_mat[0, 1] = shear_factors[0]
    shear_mat[1, 0] = shear_factors[1]
    return shear_mat


def translation_matrix(offsets: Union[Tuple[float, float], float] = (1.0, 1.0)) -> np.ndarray:
    """
    Create an Affine Transform matrix for a translation operation

    Parameters
    ----------
    offsets
        The translation ofsets.

    Returns
    -------
    A 3x3 homogeneous transform matrix implementing the translation
    """
    if isinstance(offsets, float):
        offsets = offsets, offsets
    trans_mat = np.eye(3, 3)
    trans_mat[0, 2] = offsets[1]
    trans_mat[1, 2] = offsets[0]
    return trans_mat



from typing import Generator
from typing import Iterable
from typing import Union

import numpy as np
from dkist_processing_common.models.fits_access import FitsAccessBase


def subtract_array_from_arrays(
    arrays: Union[Iterable[np.ndarray], np.ndarray],
    array_to_subtract: np.ndarray,
) -> Generator[np.ndarray, None, None]:
    """
    Subtract a single array from an iterable of arrays. This will work if
    a single array is used in lieu of an iterable as well.

    Parameters
    ----------
    arrays
        The arrays from which to subtract
    array_to_subtract
        The array to be subtracted

    Returns
    -------
    An generator of modified arrays

    """
    arrays = [arrays] if isinstance(arrays, np.ndarray) else arrays
    for array in arrays:
        if array.shape != array_to_subtract.shape:
            raise ValueError(
                f"All arrays must be the same shape. "
                f"Shape of subtraction array = {array_to_subtract.shape} "
                f"Shape of current array = {array.shape}"
            )
        yield array - array_to_subtract


def subtract_array_from_fits_access(
    access_objs: Union[Iterable[FitsAccessBase], FitsAccessBase], array_to_subtract: np.ndarray
) -> Generator[FitsAccessBase, None, None]:
    """
    Subtract a single array from an iterable of FitsAccess objects. This will work if
    a single object is used in lieu of an iterable as well. The objects themselves are modified in place.

    Parameters
    ----------
    access_objs
        The objects from which to subtract

    array_to_subtract
        The array to be subtracted

    Returns
    -------
    An generator of modified FitsAccess objects
    """
    access_objs = [access_objs] if isinstance(access_objs, FitsAccessBase) else access_objs
    for obj in access_objs:
        obj.data = next(subtract_array_from_arrays(obj.data, array_to_subtract))
        yield obj


def divide_arrays_by_array(
    arrays: Union[Iterable[np.ndarray], np.ndarray],
    array_to_divide_by: np.ndarray,
) -> Generator[np.ndarray, None, None]:
    """
    Divide an iterable of arrays by a single array. This will work if
    a single array is used in lieu of an iterable as well.

    Parameters
    ----------
    arrays
        The arrays to be divided
    array_to_divide_by
        The array ny which to divide

    Returns
    -------
    A generator of modified arrays

    """
    arrays = [arrays] if isinstance(arrays, np.ndarray) else arrays
    for array in arrays:
        if array.shape != array_to_divide_by.shape:
            raise ValueError(
                f"All arrays must be the same shape. "
                f"Shape of array to divide by = {array_to_divide_by.shape} "
                f"Shape of current array = {array.shape}"
            )
        yield array / array_to_divide_by


def divide_fits_access_by_array(
    access_objs: Union[Iterable[FitsAccessBase], FitsAccessBase], array_to_divide_by: np.ndarray
) -> Generator[FitsAccessBase, None, None]:
    """
    Divide an iterable of FitsAccess objects by a single array. This will work if
    a single object is used in lieu of an iterable as well. The objects themselves are modified in place.

    Parameters
    ----------
    access_objs
        The FitsAccess objects to be divided

    array_to_divide_by
        The array ny which to divide

    Returns
    -------
    A generator of modified FitsAccess objects
    """
    access_objs = [access_objs] if isinstance(access_objs, FitsAccessBase) else access_objs
    for obj in access_objs:
        obj.data = next(divide_arrays_by_array(obj.data, array_to_divide_by))
        yield obj



from .affine import affine_transform_arrays
from .affine import rotate_arrays_about_point
from .affine import rotation_matrix
from .affine import scale_matrix
from .affine import shear_matrix
from .affine import translate_arrays
from .affine import translation_matrix
from .binary import make_binary
from .binning import bin_arrays
from .hough import do_hough

__all__ = [
    "affine_transform_arrays",
    "translate_arrays",
    "rotate_arrays_about_point",
    "rotation_matrix",
    "translation_matrix",
    "scale_matrix",
    "shear_matrix",
    "make_binary",
    "bin_arrays",
    "do_hough",
]
